""" An interface to a database back-end for DRUIDs """

import copy
import hashlib
import logging
import os

import jsonschema
import yacman
import yaml

# module constants
# seqcol separators definition, e.g. chr1>2342>k34m6vlksb35nb,chr2>234>8h4m6vlaaab31ng
DELIM_ATTR = ">"  # separating attributes in an item (internal separator)
DELIM_ITEM = ","  # separating items in a collection (external separator)
ITEM_TYPE = "_item_type"


_LOGGER = logging.getLogger(__name__)


class NotFoundException(Exception):
    """Raised when a digest is not found"""

    def __init__(self, m):
        self.message = "{} not found in database".format(m)

    def __str__(self):
        return self.message


def md5(seq):
    return hashlib.md5(seq.encode()).hexdigest()


class Henge(object):
    def __init__(self, database, schemas, henges=None, checksum_function=md5):
        """
        A user interface to insert and retrieve decomposable recursive unique
        identifiers (DRUIDs).

        :param dict database: Dict-like lookup database for sequences and
            hashes.
        :param list schemas: One or more jsonschema schemas describing the
            data types stored by this Henge
        :param dict henges: One or more henge objects indexed by object name for
            remote storing of items.
        :param function(str) -> str checksum_function: Default function to
            handle the digest of the serialized items stored in this henge.
        """
        self.database = database
        self.checksum_function = checksum_function
        self.digest_version = checksum_function.__name__

        if isinstance(schemas, dict):
            _LOGGER.debug("Using old dict schemas")
            populated_schemas = {}
            for schema_key, schema_value in schemas.items():
                if isinstance(schema_value, str):
                    populated_schemas[schema_key] = yacman.load_yaml(schema_value)
            self.schemas = populated_schemas
        else:
            populated_schemas = []
            for schema_value in schemas:
                if isinstance(schema_value, str):
                    if os.path.isfile(schema_value):
                        populated_schemas.append(yacman.load_yaml(schema_value))
                    else:
                        populated_schemas.append(yaml.safe_load(schema_value))
            split_schemas = {}
            for s in populated_schemas:
                split_schemas.update(split_schema(s))

            self.schemas = split_schemas

        # Identify which henge to use for each item type. Default to self:
        self.henges = {}
        for item_type in self.item_types:
            self.henges[item_type] = self

        # Next add in any remote henges for item types not stored in self:
        if henges:
            for item_type, henge in henges.items():
                if item_type not in self.item_types:
                    self.schemas[item_type] = henge.schemas[item_type]
                    self.henges[item_type] = henge

    def retrieve(self, druid, reclimit=None, raw=False):
        """
        Retrieve an item given a digest

        :param str druid: The Decomposable recursive unique identifier (DRUID), or
            digest that uniquely identifies that item to retrieve.
        :param int reclimit: Recursion limit. Set to None for no limit (default).
        :param bool raw: Return the value as a raw, henge-delimited string, instead
            of processing into a mapping. Default: False.
        """

        def reconstruct_item(string, schema, reclimit):
            if "type" in schema and schema["type"] == "array":
                return [
                    reconstruct_item(substr, schema["items"], reclimit)
                    for substr in string.split(DELIM_ITEM)
                ]
            elif schema["type"] == "object":
                # else:  # assume it's an object
                attr_array = string.split(DELIM_ATTR)
                item_reconstituted = dict(zip(schema["properties"].keys(), attr_array))
                _LOGGER.debug(schema)
                if "recursive" in schema:
                    if isinstance(reclimit, int) and reclimit == 0:
                        return item_reconstituted
                    else:
                        if isinstance(reclimit, int):
                            reclimit = reclimit - 1
                        for recursive_attr in schema["recursive"]:
                            if (
                                item_reconstituted[recursive_attr]
                                and item_reconstituted[recursive_attr] != ""
                            ):
                                item_reconstituted[recursive_attr] = self.retrieve(
                                    item_reconstituted[recursive_attr], reclimit, raw
                                )
                return item_reconstituted
            else:  # it must be a primitive
                if "recursive" in schema:
                    if isinstance(reclimit, int) and reclimit == 0:
                        return string
                    else:
                        if isinstance(reclimit, int):
                            reclimit = reclimit - 1
                        return self.retrieve(string, reclimit, raw)
                else:
                    print("not recursive")
                    print(schema)
                    return string

        if not druid + ITEM_TYPE in self.database:
            raise NotFoundException(druid)

        item_type = self.database[druid + ITEM_TYPE]
        _LOGGER.debug("item_type: {}".format(item_type))
        henge_to_query = self.henges[item_type]
        _LOGGER.debug("henge_to_query: {}".format(henge_to_query))
        try:
            string = henge_to_query.database[druid]
        except KeyError:
            raise NotFoundException(druid)

        schema = self.schemas[item_type]
        return reconstruct_item(string, schema, reclimit)

    @property
    def item_types(self):
        """
        A list of item types handled by this Henge instance
        """
        return list(self.schemas.keys())

    def select_item_type(self, item):
        """
        Returns a list of all item types handled by this instance that validate
        with the given item.

        :param dict item: The item you wish to validate type of.
        """
        valid_schemas = []
        for name, schema in self.schemas.items():
            _LOGGER.debug("Testing schema: {}".format(name))
            try:
                jsonschema.validate(item, schema)
                valid_schemas.append(name)
            except jsonschema.ValidationError:
                continue
        return valid_schemas

    def insert(self, item, item_type):
        """
        Add structured items of a specified type to the database.

        :param list item: List of items to add.
        :param str item_type: A string specifying the type of item. Must match
            something from Henge.list_item_types. You can use
            Henge.select_item_type to automatically choose this, if only one
            fits.
        """

        if item_type not in self.schemas.keys():
            _LOGGER.error(
                "I don't know about items of type '{}'. "
                "I know of: '{}'".format(item_type, list(self.schemas.keys()))
            )
            return False

        schema = self.schemas[item_type]

        if not schema:
            return self.insert(item, item_type)

        if schema["type"] == "object":
            flat_item = {}
            for prop in item:
                if prop in schema["properties"]:
                    if "recursive" in schema and prop in schema["recursive"]:
                        hclass = schema["properties"][prop]["henge_class"]
                        digest = self.insert(item[prop], hclass)
                        flat_item[prop] = digest
                    else:
                        flat_item[prop] = item[prop]
                else:
                    pass  # Ignore non-schema defined properties
        elif schema["type"] == "array":
            flat_item = []
            if schema["recursive"]:
                digest = []
                hclass = schema["items"]["henge_class"]
                for element in item:
                    digest.append(self.insert(element, hclass))
                flat_item = digest
            else:
                flat_item = item

        return self._insert_flat(flat_item, item_type)

    def _insert_flat(self, item, item_type=None):
        """
        Add flattened items (of a specified type) to the database.

        Flattened items have removed all levels, so it's only attributes and
        strict values; no nesting allowed. Use the upstream insert function
        to insert full structured objects, which calls this function.

        :param list item: List of items to add.
        :param str item_type: A string specifying the type of item. Must match
            something from Henge.list_item_types. You can use
            Henge.select_item_type to automatically choose this, if only one
            fits.
        """
        if item_type not in self.schemas.keys():
            _LOGGER.error(
                "I don't know about items of type '{}'. "
                "I know of: '{}'".format(item_type, list(self.schemas.keys()))
            )
            return False

        # digest_version should be automatically appended to the item by the
        # henge. if we can put a 'default' into the schema, then the henge
        # should also populate any missing attributes with default values. can
        # jsonschema do this automatically?
        # also item_type ?

        def safestr(item, x):
            try:
                return str(item[x])
            except (ValueError, TypeError, KeyError):
                return ""

        def build_attr_string(item, schema):
            if "type" in schema and schema["type"] == "array":
                return DELIM_ITEM.join(
                    [build_attr_string(x, schema["items"]) for x in item]
                )
            elif schema["type"] == "object" and "properties" in schema:
                # else:  # assume it's an object
                return DELIM_ATTR.join(
                    [safestr(item, x) for x in list(schema["properties"].keys())]
                )
            else:  # assume it's a primitive
                return item

        valid_schema = self.schemas[item_type]
        # Add defaults here ?
        try:
            jsonschema.validate(item, valid_schema)
        except jsonschema.ValidationError as e:
            _LOGGER.error("Not valid data")
            _LOGGER.error("Attempting to insert item: {}".format(item))
            _LOGGER.error("Item type: {}".format(item_type))
            print(e)
            return False

        attr_string = build_attr_string(item, valid_schema)
        druid = self.checksum_function(attr_string)
        self._henge_insert(druid, attr_string, item_type)
        _LOGGER.debug("Loaded {}".format(druid))
        return druid

    def _henge_insert(self, druid, string, item_type, digest_version=None):
        """
        Inserts an item into the database, with henge-metadata slots for item
        type and digest version.
        """
        if not digest_version:
            digest_version = self.digest_version

        # Here we could do a few things; should we put this metadata into the
        # interface henge or the henge where the storage actually occurs? it
        # MUST be in the interface henge; should it also be in the storage
        # henge?

        henge_to_query = self.henges[item_type]
        _LOGGER.debug("henge_to_query: {}".format(henge_to_query))
        try:
            henge_to_query.database[druid] = string
            henge_to_query.database[druid + ITEM_TYPE] = item_type
            henge_to_query.database[druid + "_digest_version"] = digest_version

            if henge_to_query != self:
                self.database[druid + ITEM_TYPE] = item_type
                self.database[druid + "_digest_version"] = digest_version
        except Exception as e:
            raise e

    def clean(self):
        """
        Remove all items from this database.
        """
        for k, v in self.database.items():
            try:
                del self.database[k]
                del self.database[k + ITEM_TYPE]
                del self.database[k + "_digest_version"]
            except (KeyError, AttributeError):
                pass

    def show(self):
        """
        Show all items in the database.
        """
        for k, v in self.database.items():
            print(k, v)

    def __repr__(self):
        repr = (
            "Henge object\n"
            + "Item types: "
            + ",".join(self.item_types)
            + "\n"
            + "Schemas: "
            + str(self.schemas)
        )
        return repr


def split_schema(schema, name=None):
    """
    Splits a hierarchical schema into flat components suitable for a Henge
    """
    slist = {}
    # base case
    if schema["type"] not in ["object", "array"]:
        _LOGGER.debug(schema)
        if name:
            slist[name] = schema
        elif "henge_class" in schema:
            slist[schema["henge_class"]] = schema
        _LOGGER.debug("Returning slist: {}".format(str(slist)))
        return slist
    elif schema["type"] == "object":
        if "henge_class" in schema:
            schema_copy = copy.deepcopy(schema)
            _LOGGER.debug("adding " + str(schema_copy["henge_class"]))
            henge_class = schema_copy["henge_class"]
            # del schema_copy['henge_class']
            for p in schema_copy["properties"]:
                hclass = None
                if "henge_class" in schema_copy["properties"][p]:
                    hclass = schema_copy["properties"][p]["henge_class"]
                if schema_copy["properties"][p]["type"] in ["object", "array"]:
                    schema_copy["properties"][p] = {"type": "string"}
                    if hclass:
                        schema_copy["properties"][p][
                            "henge_class"
                        ] = hclass  # schema_copy['properties'][p]['type'] = "string"
            # del schema_copy['properties']
            slist[henge_class] = schema_copy

        for p in schema["properties"]:
            schema_sub = schema["properties"][p]
            _LOGGER.debug("checking property:" + p)
            slist.update(split_schema(schema["properties"][p]))
    elif schema["type"] == "array":
        _LOGGER.debug("found array")
        if "henge_class" in schema:
            schema_copy = copy.deepcopy(schema)
            _LOGGER.debug("adding", schema_copy["henge_class"])
            henge_class = schema_copy["henge_class"]
            # del schema_copy['henge_class']
            schema_copy["items"] = {"type": "string"}
            if "recursive" in schema_copy:
                schema_copy["items"]["recursive"] = True
            if "henge_class" in schema["items"]:
                schema_copy["items"]["henge_class"] = schema["items"]["henge_class"]
            # schema_copy['items']['type'] = "string"
            # if 'properties' in schema_copy['items']:
            #     del schema_copy['items']['properties']
            slist[henge_class] = schema_copy

        schema_sub = schema["items"]
        _LOGGER.debug("Checking item")
        slist.update(split_schema(schema_sub))
    return slist



# refgenie looper plugin

import logging
from collections.abc import Mapping

from attmap import AttMap
from ubiquerg import parse_registry_path as prp

import refgenconf

_LOGGER = logging.getLogger(__name__)


def looper_refgenie_populate(namespaces):
    """
    A looper plugin that populates refgenie references in a PEP from
    refgenie://genome/asset:tag registry paths. This can be used to convert
    all refgenie references into their local paths at the looper stage, so the
    final paths are passed to the workflow. This way the workflow does not
    need to depend on refgenie to resolve the paths.
    This is useful for example for CWL pipelines, which are built to have
    paths resolved outside the workflow.

    The namespaces structure required to run the plugin is:
    `namespaces["pipeline"]["var_templates"]["refgenie_config"]`

    :param Mapping namespaces: a nested variable namespaces dict
    :return dict: sample namespace dict
    :raises TypeError: if the input namespaces is not a mapping
    :raises KeyError: if the namespaces mapping does not include 'pipeline'
    :raises NotImplementedError: if 'var_templates' key is missing in the 'pipeline' namespace or
        'refgenie_config' is missing in 'var_templates' section.
    """
    if not isinstance(namespaces, Mapping):
        raise TypeError("Namespaces must be a Mapping")
    if "pipeline" not in namespaces:
        raise KeyError(
            "Namespaces do not include 'pipeline'. The job is misconfigured."
        )
    if (
        "var_templates" in namespaces["pipeline"]
        and "refgenie_config" in namespaces["pipeline"]["var_templates"]
    ):
        rgc_path = namespaces["pipeline"]["var_templates"]["refgenie_config"]
        rgc = refgenconf.RefGenConf(rgc_path)

        complete_sk_dict = rgc.list_seek_keys_values()
        paths_dict = {}

        # This function allows you to specify tags for specific assets to use
        # in the project config like:
        # refgenie_asset_tags:
        #   genome:
        #     asset_name: tag_name
        def get_asset_tag(genome, asset):
            try:
                return namespaces["project"]["refgenie"]["tag_overrides"][genome][asset]
            except KeyError:
                default_tag = rgc.get_default_tag(genome=genome, asset=asset)
                _LOGGER.info(
                    f"Refgenie asset ({genome}/{asset}) tag not specified in `refgenie.tag_overrides` section. "
                    f"Using the default tag: {default_tag}"
                )
                return default_tag
            except TypeError:
                default_tag = rgc.get_default_tag(genome=genome, asset=asset)
                _LOGGER.warn(f"tag_overrides section is malformed. Using default.")
                return default_tag

        # Restructure the seek key paths to make them accessible with
        # {refgenie.asset_name.seek_key} in command templates
        for g, gdict in complete_sk_dict.items():
            _LOGGER.debug(f"Processing genome {g}")
            paths_dict[g] = {}
            for k, v in gdict.items():
                tag = get_asset_tag(genome=g, asset=k)
                # print(k,v)
                try:
                    paths_dict[g][k] = v[tag]
                except KeyError:
                    _LOGGER.warn(
                        f"Can't find tag '{tag}' for asset '{g}/{k}', as specified in your project config. Using default."
                    )
                    paths_dict[g][k] = v[rgc.get_default_tag(genome=g, asset=k)]

        if "project" in namespaces and "refgenie" in namespaces["project"]:
            try:
                for po in namespaces["project"]["refgenie"]["path_overrides"]:
                    rp = prp(po["registry_path"])
                    _LOGGER.debug(
                        f"Overriding {po['registry_path']} with {po['value']}."
                    )
                    if not rp["subitem"]:
                        rp["subitem"] = rp["item"]
                    _LOGGER.debug(rp)
                    paths_dict[rp["namespace"]][rp["item"]][rp["subitem"]] = po["value"]
            except KeyError:
                _LOGGER.debug("Did not find path_overrides section")
            except TypeError:
                _LOGGER.warn("Warning: path_overrides is not iterable")

        # print(paths_dict)
        # Provide these values under the 'refgenie' namespace
        namespaces["refgenie"] = AttMap(paths_dict)
        return rgc.populate(namespaces)
    else:
        msg = """
        var_templates:
          refgenie_config: "$REFGENIE"
        """
        _LOGGER.error(
            f"refgenie_config not specified in pipeline interface. Do like so: {msg}"
        )
        raise NotImplementedError



""" Package exception types """

import abc

__all__ = [
    "DownloadJsonError",
    "GenomeConfigFormatError",
    "MissingAssetError",
    "MissingRecipeError",
    "MissingConfigDataError",
    "MissingGenomeError",
    "MissingSeekKeyError",
    "MissingTagError",
    "RefgenconfError",
    "UnboundEnvironmentVariablesError",
    "ConfigNotCompliantError",
    "RemoteDigestMismatchError",
]

DOC_URL = "http://refgenie.databio.org/en/latest/genome_config/"


class RefgenconfError(Exception):
    """Base exception type for this package"""

    __metaclass__ = abc.ABCMeta


class DownloadJsonError(RefgenconfError):
    """Non-OK response from a JSON download attempt"""

    def __init__(self, resp):
        super(DownloadJsonError, self).__init__(
            "No response provided" if resp is None else "JSON: {}".format(resp.json())
        )
        self.response = resp


class GenomeConfigFormatError(RefgenconfError):
    """Exception for invalid genome config file format."""

    def __init__(self, msg):
        spacing = " " if msg[-1] in ["?", ".", "\n"] else "; "
        suggest = "For config format documentation please see " + DOC_URL
        super(GenomeConfigFormatError, self).__init__(msg + spacing + suggest)


class MissingAssetError(RefgenconfError):
    """Error type for request of an unavailable genome asset."""

    pass


class MissingTagError(RefgenconfError):
    """Error type for request of an unavailable asset tag."""

    pass


class MissingSeekKeyError(RefgenconfError):
    """Error type for request of an unavailable asset seek key."""

    pass


class MissingRecipeError(RefgenconfError):
    """Error type for request of an unavailable recipe."""

    pass


class MissingConfigDataError(RefgenconfError):
    """Missing required configuration instance items"""

    pass


class ConfigNotCompliantError(GenomeConfigFormatError):
    """The format of the config file does not match required version/standards"""

    pass


class MissingGenomeError(RefgenconfError):
    """Error type for request of unknown genome/assembly."""

    pass


class UnboundEnvironmentVariablesError(RefgenconfError):
    """Use of environment variable that isn't bound to a value."""

    pass


class RemoteDigestMismatchError(RefgenconfError):
    """Remote digest of the parent asset does not match its local counterpart"""

    def __init__(self, asset, local_digest, remote_digest):
        msg = (
            "This asset is built from parent asset '{}', but for this parent, the remote does not "
            "match your local asset (local: {}; remote: {}). Refgenie will not pull this asset "
            "because the remote version was not built from the same parent asset you have locally.".format(
                asset, local_digest, remote_digest
            )
        )
        super(RemoteDigestMismatchError, self).__init__(msg)



"""
Config file structure determination for the refgenie suite of packages

These values are defined here in refgenconf and use some within this package,
but they're also integral to both refgenie and to refgenieserver.
"""

import os

CFG_NAME = "genome configuration"
CFG_ENV_VARS = ["REFGENIE"]
CFG_CONST = ["CFG_NAME", "CFG_ENV_VARS"]
DEFAULT_SERVER = "http://refgenomes.databio.org"  # "http://rg.databio.org"
TAG_NAME_BANNED_CHARS = [":", "/"]
API_VERSION = "v3"
API_VERSION_2 = "v2"
DEFAULT_TAG = "default"
DEFAULT_CONFIG_SCHEMA = os.path.join(
    os.path.dirname(__file__), "schemas", "genome_config_schema.yaml"
)

# file or dir names
TEMPLATE_RECIPE_JSON = "build_recipe_{}__{}.json"
TEMPLATE_TARGET = "{}_{}__{}.flag"
TEMPLATE_LOG = "build_log_{}__{}.md"
TEMPLATE_ASSET_DIR_CONTENTS = "asset_dir_contents_{}__{}.json"
ORI_LOG_NAME = "refgenie_log.md"
BUILD_STATS_DIR = "_refgenie_build"
LOCKED_BUILD_MAP_CFG = "_locked_map_build.yaml"
BUILD_MAP_CFG = "_map_build.yaml"
ALIAS_DIR = "alias"
DATA_DIR = "data"

FILE_DIR_NAMES = [
    "TEMPLATE_RECIPE_JSON",
    "TEMPLATE_TARGET",
    "TEMPLATE_LOG",
    "TEMPLATE_ASSET_DIR_CONTENTS",
    "ORI_LOG_NAME",
    "BUILD_STATS_DIR",
    "BUILD_MAP_CFG",
    "LOCKED_BUILD_MAP_CFG",
    "ALIAS_DIR",
    "DATA_DIR",
]

# project-wide definition of the endpoint IDs. They are used to establish the
# way of communication between the server and the client so that changes of
# endpoint function names OR endpoints themselves do not influence the connection
CUSTOM_PFX = "custom_Id"
API_ID_ALIAS_ALIAS = CUSTOM_PFX + "_alias_alias"
API_ID_ALIAS_DIGEST = CUSTOM_PFX + "_alias_digest"
API_ID_ALIASES_DICT = CUSTOM_PFX + "_aliases_dict"
API_ID_ASSETS = CUSTOM_PFX + "_assets"
API_ID_ARCHIVE = CUSTOM_PFX + "_archive"
API_ID_DEFAULT_TAG = CUSTOM_PFX + "_default_tag"
API_ID_ASSET_ATTRS = CUSTOM_PFX + "_asset_attrs"
API_ID_GENOME_ATTRS = CUSTOM_PFX + "_genome_attrs"
API_ID_DIGEST = CUSTOM_PFX + "_asset_digest"
API_ID_RECIPE = CUSTOM_PFX + "_asset_recipe"
API_ID_LOG = CUSTOM_PFX + "_asset_log"
API_ID_ASSET_FILE = CUSTOM_PFX + "_asset_file"
API_ID_ASSET_PATH = CUSTOM_PFX + "_asset_path"
API_ID_ARCHIVE_DIGEST = CUSTOM_PFX + "_asset_archive_digest"
API_ID_SPLASH = CUSTOM_PFX + "_asset_splash"
API_ID_GENOMES_DICT = CUSTOM_PFX + "_genomes_dict"
API_ID_CONTENTS = CUSTOM_PFX + "_asset_dir_contents"

PRIVATE_API = "_private_api"

# this dictionary groups the operationIds so that they can be accessed as
# modules for systematic links generation in the splash pages
OPERATION_IDS = {
    "asset": {
        API_ID_ARCHIVE: "archive",
        API_ID_ASSET_ATTRS: "attributes",
        API_ID_DIGEST: "asset digest",
        API_ID_ARCHIVE_DIGEST: "archive digest",
        API_ID_RECIPE: "build recipe",
        API_ID_LOG: "build log",
    },
    "v3_asset": {
        API_VERSION + API_ID_ARCHIVE: "archive",
        API_VERSION + API_ID_ASSET_ATTRS: "attributes",
        API_VERSION + API_ID_DIGEST: "asset digest",
        API_VERSION + API_ID_ARCHIVE_DIGEST: "archive digest",
        API_VERSION + API_ID_RECIPE: "build recipe",
        API_VERSION + API_ID_LOG: "build log",
        API_VERSION + API_ID_CONTENTS: "asset directory contents",
    },
}

API_IDS = [
    "API_ID_ASSETS",
    "API_ID_ARCHIVE",
    "API_ID_DEFAULT_TAG",
    "API_ID_LOG",
    "API_ID_CONTENTS",
    "API_ID_ASSET_FILE",
    "API_ID_ASSET_PATH",
    "API_ID_DIGEST",
    "API_ID_RECIPE",
    "API_ID_ASSET_ATTRS",
    "API_ID_SPLASH",
    "API_ID_ALIASES_DICT",
    "API_ID_ARCHIVE_DIGEST",
    "API_ID_ALIAS_ALIAS",
    "API_ID_ALIAS_DIGEST",
    "API_ID_GENOME_ATTRS",
    "API_ID_GENOMES_DICT",
]

CFG_FOLDER_KEY = "genome_folder"
CFG_SERVERS_KEY = "genome_servers"
CFG_SERVER_KEY = "genome_server"
CFG_ARCHIVE_KEY = "genome_archive_folder"
CFG_ARCHIVE_KEY_OLD = "genome_archive"
CFG_ARCHIVE_CONFIG_KEY = "genome_archive_config"
CFG_REMOTE_URL_BASE_KEY = "remote_url_base"
CFG_VERSION_KEY = "config_version"
CFG_GENOMES_KEY = "genomes"
CFG_ALIASES_KEY = "aliases"

CFG_CHECKSUM_KEY = "genome_digest"
CFG_GENOME_DESC_KEY = "genome_description"
CFG_ASSETS_KEY = "assets"

CFG_GENOME_MASK_KEY = "genome_mask"
CFG_ASSET_PATH_KEY = "asset_path"
CFG_ASSET_SIZE_KEY = "asset_size"
CFG_ASSET_DESC_KEY = "asset_description"
CFG_ARCHIVE_SIZE_KEY = "archive_size"
CFG_ARCHIVE_CHECKSUM_KEY = "archive_digest"
CFG_SEEK_KEYS_KEY = "seek_keys"
CFG_ASSET_PARENTS_KEY = "asset_parents"
CFG_ASSET_CHILDREN_KEY = "asset_children"
CFG_ASSET_DEFAULT_TAG_KEY = "default_tag"
CFG_ASSET_TAGS_KEY = "tags"
CFG_ASSET_CHECKSUM_KEY = "asset_digest"
CFG_TAG_DESC_KEY = "tag_description"

CFG_ASSET_RELATIVES_KEYS = [CFG_ASSET_CHILDREN_KEY, CFG_ASSET_PARENTS_KEY]

CFG_TOP_LEVEL_KEYS = [
    CFG_FOLDER_KEY,
    CFG_SERVER_KEY,
    CFG_SERVERS_KEY,
    CFG_ARCHIVE_KEY,
    CFG_GENOMES_KEY,
    CFG_ALIASES_KEY,
    CFG_VERSION_KEY,
    CFG_ARCHIVE_CONFIG_KEY,
    CFG_ARCHIVE_KEY_OLD,
    CFG_REMOTE_URL_BASE_KEY,
]
CFG_GENOME_KEYS = [CFG_GENOME_DESC_KEY, CFG_ASSETS_KEY, CFG_CHECKSUM_KEY]
CFG_GENOME_ATTRS_KEYS = [CFG_GENOME_DESC_KEY, CFG_CHECKSUM_KEY]
CFG_SINGLE_ASSET_SECTION_KEYS = [
    CFG_ASSET_PATH_KEY,
    CFG_ASSET_DESC_KEY,
    CFG_ASSET_SIZE_KEY,
    CFG_ARCHIVE_SIZE_KEY,
    CFG_ARCHIVE_CHECKSUM_KEY,
    CFG_SEEK_KEYS_KEY,
    CFG_GENOME_MASK_KEY,
]

RGC_REQ_KEYS = [CFG_SERVERS_KEY, CFG_FOLDER_KEY, CFG_GENOMES_KEY, CFG_VERSION_KEY]

CFG_KEY_NAMES = [
    "CFG_FOLDER_KEY",
    "CFG_SERVER_KEY",
    "CFG_SERVERS_KEY",
    "CFG_GENOMES_KEY",
    "CFG_GENOME_MASK_KEY",
    "CFG_ALIASES_KEY",
    "CFG_ASSET_PATH_KEY",
    "CFG_ASSET_DESC_KEY",
    "CFG_ARCHIVE_KEY",
    "CFG_ARCHIVE_SIZE_KEY",
    "CFG_SEEK_KEYS_KEY",
    "CFG_ASSET_SIZE_KEY",
    "CFG_CHECKSUM_KEY",
    "CFG_ARCHIVE_CHECKSUM_KEY",
    "CFG_VERSION_KEY",
    "CFG_ASSET_PARENTS_KEY",
    "CFG_ASSET_CHILDREN_KEY",
    "CFG_TAG_DESC_KEY",
    "CFG_ASSET_CHECKSUM_KEY",
    "CFG_ASSET_TAGS_KEY",
    "CFG_ASSET_RELATIVES_KEYS",
    "CFG_ARCHIVE_CONFIG_KEY",
    "CFG_ARCHIVE_KEY_OLD",
    "CFG_REMOTE_URL_BASE_KEY",
]

# hook identifiers
PRE_UPDATE_HOOK = "pre_update"
POST_UPDATE_HOOK = "post_update"
PRE_PULL_HOOK = "pre_pull"
POST_PULL_HOOK = "post_pull"
PRE_TAG_HOOK = "pre_tag"
POST_TAG_HOOK = "post_tag"
PRE_LIST_HOOK = "pre_list"
POST_LIST_HOOK = "post_list"
# HOOKS is a list of all available plugin entry points
HOOK_NAMES = [
    "PRE_LIST_HOOK",
    "PRE_PULL_HOOK",
    "PRE_TAG_HOOK",
    "PRE_UPDATE_HOOK",
    "POST_TAG_HOOK",
    "POST_LIST_HOOK",
    "POST_PULL_HOOK",
    "POST_UPDATE_HOOK",
]
HOOKS = [eval(x) for x in HOOK_NAMES]

# other consts
REQ_CFG_VERSION = 0.4
REFGENIE_BY_CFG = {"0.4": "0.10.0", "0.3": "0.7.0", "0.2": "0.6.0"}
CFG_UPGRADE = {"0.3": ["0.4"]}
ATTRS_COPY_PULL = [
    CFG_ASSET_DESC_KEY,
    CFG_SEEK_KEYS_KEY,
    CFG_ASSET_PARENTS_KEY,
    CFG_ASSET_PATH_KEY,
    CFG_ASSET_CHECKSUM_KEY,
    CFG_TAG_DESC_KEY,
]
REQ_TAG_ATTRS = [CFG_ASSET_PATH_KEY, CFG_SEEK_KEYS_KEY]
CUSTOM_BAR_FMT = "{desc}{percentage:3.0f}%|{bar}| {n_fmt} [{elapsed}<{remaining} {rate_fmt}{postfix}]"

__all__ = (
    [
        "DEFAULT_SERVER",
        "CFG_ASSET_DEFAULT_TAG_KEY",
        "CFG_KEY_NAMES",
        "CFG_GENOME_DESC_KEY",
        "REQ_CFG_VERSION",
        "CFG_ASSETS_KEY",
        "CFG_GENOME_ATTRS_KEYS",
        "REFGENIE_BY_CFG",
        "CFG_UPGRADE",
        "DEFAULT_TAG",
        "ATTRS_COPY_PULL",
        "RGC_REQ_KEYS",
        "REQ_TAG_ATTRS",
        "CUSTOM_BAR_FMT",
        "API_VERSION",
        "API_VERSION_2",
        "TAG_NAME_BANNED_CHARS",
        "CONF_STRUCTURE",
        "OPERATION_IDS",
        "CUSTOM_PFX",
        "PRIVATE_API",
        "HOOKS",
        "DEFAULT_CONFIG_SCHEMA",
    ]
    + FILE_DIR_NAMES
    + CFG_CONST
    + CFG_KEY_NAMES
    + API_IDS
    + HOOK_NAMES
)

CONF_STRUCTURE = """
# example genome configuration structure
{version}: {v}
{folder}: $GENOMES
{server}: http://localhost
{archive}: /path/to/archives

{genomes}:
    fcdd62cb90e86d03e45dcd05efa70d8bdc9577d5c6259cf5:
        {aliases}: ['hg38']
        {desc_genome}: Reference assembly GRCh38, released in Dec 2013
        {assets}:
            fasta:
                {default}: tag_name
                {desc_asset}: DNA sequences in the FASTA format, indexed FASTA (produced with samtools index) and chromosome sizes file
                {tags}:
                    tag_name:
                        {asset_path}: fasta
                        {archive_digest}: 35ae9a42c36c126f9d8ef6d938a122d0
                        {asset_digest}: 3aff393d290884336945534ea709d30e
                        {asset_size}: 3.0GB
                        {archive_size}: 938.3MB
                        {asset_parents}:[]
                        {asset_children}: []
                        {seek_keys}:
                            fasta: fcdd62cb90e86d03e45dcd05efa70d8bdc9577d5c6259cf5.fa.gz
                            fai: fcdd62cb90e86d03e45dcd05efa70d8bdc9577d5c6259cf5.fa.fai
                            chrom_sizes: fcdd62cb90e86d03e45dcd05efa70d8bdc9577d5c6259cf5.chrom.sizes
""".format(
    folder=CFG_FOLDER_KEY,
    server=CFG_SERVERS_KEY,
    version=CFG_VERSION_KEY,
    assets=CFG_ASSETS_KEY,
    archive=CFG_ARCHIVE_KEY,
    digest=CFG_CHECKSUM_KEY,
    genomes=CFG_GENOMES_KEY,
    aliases=CFG_ALIASES_KEY,
    desc_genome=CFG_GENOME_DESC_KEY,
    asset_path=CFG_ASSET_PATH_KEY,
    desc_asset=CFG_ASSET_DESC_KEY,
    archive_digest=CFG_ARCHIVE_CHECKSUM_KEY,
    asset_size=CFG_ASSET_SIZE_KEY,
    archive_size=CFG_ARCHIVE_SIZE_KEY,
    seek_keys=CFG_SEEK_KEYS_KEY,
    asset_parents=CFG_ASSET_PARENTS_KEY,
    asset_children=CFG_ASSET_CHILDREN_KEY,
    default=CFG_ASSET_DEFAULT_TAG_KEY,
    tags=CFG_ASSET_TAGS_KEY,
    asset_digest=CFG_ASSET_CHECKSUM_KEY,
    tag_description=CFG_TAG_DESC_KEY,
    v=REQ_CFG_VERSION,
)



import binascii
import hashlib
import logging
import os
from gzip import open as gzopen

from .exceptions import RefgenconfError
from .henge import ITEM_TYPE, Henge


def trunc512_digest(seq, offset=24):
    digest = hashlib.sha512(seq.encode()).digest()
    hex_digest = binascii.hexlify(digest[:offset])
    return hex_digest.decode()


# module constants
def _schema_path(name):
    return os.path.join(SCHEMA_FILEPATH, name)


CONTENT_ALL_A_IN_B = 2 ** 0
CONTENT_ALL_B_IN_A = 2 ** 1
LENGTHS_ALL_A_IN_B = 2 ** 2
LENGTHS_ALL_B_IN_A = 2 ** 3
NAMES_ALL_A_IN_B = 2 ** 4
NAMES_ALL_B_IN_A = 2 ** 5
CONTENT_A_ORDER = 2 ** 6
CONTENT_B_ORDER = 2 ** 7
CONTENT_ANY_SHARED = 2 ** 8
LENGTHS_ANY_SHARED = 2 ** 9
NAMES_ANY_SHARED = 2 ** 10

FLAGS = {
    CONTENT_ALL_A_IN_B: "CONTENT_ALL_A_IN_B",
    CONTENT_ALL_B_IN_A: "CONTENT_ALL_B_IN_A",
    LENGTHS_ALL_A_IN_B: "LENGTHS_ALL_A_IN_B",
    LENGTHS_ALL_B_IN_A: "LENGTHS_ALL_B_IN_A",
    NAMES_ALL_A_IN_B: "NAMES_ALL_A_IN_B",
    NAMES_ALL_B_IN_A: "NAMES_ALL_B_IN_A",
    CONTENT_ANY_SHARED: "CONTENT_ANY_SHARED",
    LENGTHS_ANY_SHARED: "LENGTHS_ANY_SHARED",
    NAMES_ANY_SHARED: "NAMES_ANY_SHARED",
    CONTENT_A_ORDER: "CONTENT_A_ORDER",
    CONTENT_B_ORDER: "CONTENT_B_ORDER",
}

NAME_KEY = "name"
SEQ_KEY = "sequence"
LEN_KEY = "length"

# internal schemas paths determination
ASL_NAME = "AnnotatedSequenceList"
ASDL_NAME = "AnnotatedSequenceDigestList"
SCHEMA_NAMES = [ASL_NAME, ASDL_NAME]
SCHEMA_FILEPATH = os.path.join(os.path.dirname(__file__), "schemas")
INTERNAL_SCHEMAS = [_schema_path(f"{s}.yaml") for s in SCHEMA_NAMES]

_LOGGER = logging.getLogger(__name__)


class SeqColClient(Henge):
    """
    Extension of henge that accommodates collections of sequences.
    """

    def __init__(
        self, database, schemas=None, henges=None, checksum_function=trunc512_digest
    ):
        """
        A user interface to insert and retrieve decomposable recursive unique
        identifiers (DRUIDs).

        :param dict database: Dict-like lookup database with sequences
            and hashes
        :param dict schemas: One or more jsonschema schemas describing the
            data types stored by this Henge
        :param function(str) -> str checksum_function: Default function to
            handle the digest of the
            serialized items stored in this henge.
        """
        assert all([os.path.exists(s) for s in INTERNAL_SCHEMAS]), RefgenconfError(
            f"Missing schema files: {INTERNAL_SCHEMAS}"
        )
        super(SeqColClient, self).__init__(
            database=database,
            schemas=schemas or INTERNAL_SCHEMAS,
            henges=henges,
            checksum_function=checksum_function,
        )

    def load_fasta(self, fa_file, skip_seq=False, gzipped=False):
        """
        Load a sequence collection into the database

        :param str fa_file: path to the FASTA file to parse and load
        :param bool skip_seq: whether to disregard the actual sequences,
            load just the names and lengths
        :param bool skip_seq: whether to disregard the actual sequences,
            load just the names and lengths
        """
        seq = ""
        name = ""
        init = False
        aslist = []
        openfun = gzopen if gzipped else open
        with openfun(fa_file, "rt") as f:
            for line in f:
                line = line.strip("\n")
                if line.startswith(">"):
                    if not init:
                        name = line.replace(">", "")
                    else:
                        aslist.append(
                            {
                                NAME_KEY: name,
                                LEN_KEY: len(seq),
                                SEQ_KEY: "" if skip_seq else trunc512_digest(seq),
                            }
                        )
                        name = line.replace(">", "")
                    seq = ""
                    continue
                init = True
                seq = seq + line
            aslist.append(
                {
                    NAME_KEY: name,
                    LEN_KEY: len(seq),
                    SEQ_KEY: "" if skip_seq else trunc512_digest(seq),
                }
            )

        collection_checksum = self.insert(aslist, ASDL_NAME)
        _LOGGER.info(f"Loaded {ASDL_NAME} ({len(aslist)} sequences)")
        return collection_checksum, aslist

    @staticmethod
    def compare_asds(asdA, asdB, explain=False):
        """
        Compare Annotated Sequence Digests (ASDs) -- digested sequences and metadata

        :param str asdA: ASD for first sequence collection to compare.
        :param str asdB: ASD for second sequence collection to compare.
        :param bool explain: Print an explanation of the flag? [Default: False]
        """

        def _xp(prop, lst):
            """Extract property from a list of dicts"""
            return list(map(lambda x: x[prop], lst))

        def _index(x, lst):
            """Find an index of a sequence element in a list of dicts"""
            try:
                return _xp(SEQ_KEY, lst).index(x)
            except:
                return None

        def _get_common_content(lstA, lstB):
            """
            Find the intersection between two list of dicts with sequences
            """
            return list(
                filter(None.__ne__, [_index(x, lstB) for x in _xp(SEQ_KEY, lstA)])
            )

        # Not ideal, but we expect these to return lists, but if the item was
        # singular only a dict is returned
        if not isinstance(asdA, list):
            asdA = [asdA]
        if not isinstance(asdB, list):
            asdB = [asdB]

        ainb = [x in _xp(SEQ_KEY, asdB) for x in _xp(SEQ_KEY, asdA)]
        bina = [x in _xp(SEQ_KEY, asdA) for x in _xp(SEQ_KEY, asdB)]

        return_flag = 0  # initialize
        if any(ainb):
            ordA = _get_common_content(asdA, asdB)
            if ordA == sorted(ordA):
                return_flag += CONTENT_A_ORDER
        if any(bina):
            ordB = _get_common_content(asdB, asdA)
            if ordB == sorted(ordB):
                return_flag += CONTENT_B_ORDER

        ainb_len = [x in _xp(LEN_KEY, asdB) for x in _xp(LEN_KEY, asdA)]
        bina_len = [x in _xp(LEN_KEY, asdA) for x in _xp(LEN_KEY, asdB)]

        ainb_name = [x in _xp(NAME_KEY, asdB) for x in _xp(NAME_KEY, asdA)]
        bina_name = [x in _xp(NAME_KEY, asdA) for x in _xp(NAME_KEY, asdB)]

        if any(ainb):
            return_flag += CONTENT_ANY_SHARED
        if all(ainb):
            return_flag += CONTENT_ALL_A_IN_B
        if all(bina):
            return_flag += CONTENT_ALL_B_IN_A

        if any(ainb_name):
            return_flag += NAMES_ANY_SHARED
        if all(ainb_name):
            return_flag += NAMES_ALL_A_IN_B
        if all(bina_name):
            return_flag += NAMES_ALL_B_IN_A

        if any(ainb_len):
            return_flag += LENGTHS_ANY_SHARED
        if all(ainb_len):
            return_flag += LENGTHS_ALL_A_IN_B
        if all(bina_len):
            return_flag += LENGTHS_ALL_B_IN_A

        if explain:
            explain_flag(return_flag)
        return return_flag

    def compare(self, digestA, digestB, explain=False):
        """
        Given two collection checksums in the database, provide some information
        about how they are related.

        :param str digestA: Digest for first sequence collection to compare.
        :param str digestB: Digest for second sequence collection to compare.
        :param bool explain: Print an explanation of the flag? [Default: False]
        """
        typeA = self.database[digestA + ITEM_TYPE]
        typeB = self.database[digestB + ITEM_TYPE]

        if typeA != typeB:
            _LOGGER.error(
                f"Can't compare objects of different types: " f"{typeA} vs {typeB}"
            )

        asdA = self.retrieve(digestA, reclimit=1)
        asdB = self.retrieve(digestB, reclimit=1)
        return self.compare_asds(asdA, asdB, explain=explain)


# Static functions below (these don't require a database)


def explain_flag(flag):
    """Explains a compare flag"""
    print(f"Flag: {flag}\nBinary: {bin(flag)}\n")
    for e in range(0, 13):
        if flag & 2 ** e:
            print(FLAGS[2 ** e])
